#!/usr/bin/env python
"""
  cbmls is an ls-like file lister for Commodore emulator disk/file images
  (like .d64 and .p00) and archive files (like .lnx and [1-4]! zip files).
  Supports most formats, recursive listings, deep scans, and more.
"""

# customize your own preferred defaults here, or comment it out
arg_defaults = [('almost_all', True), ('long', True), ('size', True), ('unicode', False), ('color', True)]


# standard imports necessary for finding the terminal width
import shlex
import struct
import platform

# standard imports necessary for everything else
import argparse
import glob
import os
import string
import subprocess
import sys
from math import log
from operator import attrgetter

# local import for CBMDisk
import cbmdisk


def cli():
    """ handle command line arguments """
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog='using CBMdisk/python v' + cbmdisk.__version__ + ' w/ CBMdisk v' + cbmdisk.library_version
                                     )
    parser.add_argument('filename', nargs='+', metavar='FILE',
                        help='a disk/file image or archive filename')
    parser.add_argument('--version', action='version', version='%(prog)s v1.0.0; by Elwix/Style; terminal width code: Harco Kuppens, jtriley',
                        help='show program version')
    parser.add_argument('-1', dest='one', action='store_true',
                        help='list one file per line')
    parser.add_argument('-a', '--all', action='store_true',
                        help='list all file entries')
    parser.add_argument('-H', '--human-readable', action='store_true',
                        help='with -s, print byte sizes in human readable format')
    parser.add_argument('-i', '--index', action='store_true',
                        help='print the index number of each file')
    parser.add_argument('-l', dest='long', action='store_true',
                        help='use a long listing format')
    parser.add_argument('-q', '--quiet', action='count',
                        help='suppress decompress errors (-q) and disk errors (-qq)')
    parser.add_argument('-r', '--reverse', action='store_true',
                        help='reverse order while sorting')
    parser.add_argument('-s', '--size', action='store_true',
                        help='print size of each file, in blocks and bytes')
    parser.add_argument('-A', '--almost-all', action='store_true',
                        help='list all but apparently unallocated file entries')
    parser.add_argument('-C', '--case', choices=['upper', 'lower'], default='upper',
                        help='output as if in upper/graphics or lower/upper mode')
    parser.add_argument('-D', '--deep', action='store_true',
                        help='list cbm subdirs and archives recursively')
    parser.add_argument('-N', '--sort-name', action='store_true',
                        help='sort alphabetically by file name')
    parser.add_argument('-R', '--recursive', action='store_true',
                        help='list subdirectories recursively')
    parser.add_argument('-S', '--sort-size', action='store_true',
                        help='sort numerically by file size')
    parser.add_argument('-U', '--unicode', action='store_true', default=False,
                        help='output contents as unicode (from petscii) instead of ascii')
    parser.add_argument('-X', '--sort-type', action='store_true',
                        help='sort alphabetically by file type')
    parser.add_argument('--color', action='store_true',
                        help='colorize the output')
    try:
        parser.set_defaults(**dict(arg_defaults))
    except NameError:
        pass
    ret_args = parser.parse_args()
    return ret_args


# list of suffixes that we expect to be able to open and read as an archival format
archive_suffix = ('lnx', 'arc', 'sda', 'sfx', 'wr3', 'lbr', 'ark', 'srk', 'spy', 'cvt')

# list of prefixes that we expect to be able to open and read as an archival format
archive_prefix = ('1!', '2!', '3!', '4!')


# terminal code from https://gist.github.com/jtriley/1108174 (and originally from https://stackoverflow.com/a/6550596)
# (lots of code, needed for python 2.x)
def get_terminal_size():
    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!
    if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
        tuple_xy = _get_terminal_size_linux()
    if tuple_xy is None:
        print "default"
        tuple_xy = (40, 25)      # default value
    return tuple_xy


if platform.system() == "Windows":
    def _get_terminal_size_windows():
        try:
            from ctypes import windll, create_string_buffer
            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12
            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (bufx, bufy, curx, cury, wattr,
                 left, top, right, bottom,
                 maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1
                sizey = bottom - top + 1
                return sizex, sizey
        except:
            pass

    def _get_terminal_size_tput():
        # get terminal width
        # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
        try:
            cols = int(subprocess.check_call(shlex.split('tput cols')))
            rows = int(subprocess.check_call(shlex.split('tput lines')))
            return cols, rows
        except:
            pass


if platform.system() in ['Linux', 'Darwin'] or platform.system().startswith('CYGWIN'):
    def _get_terminal_size_linux():
        # noinspection PyUnresolvedReferences
        def ioctl_gwinsz(fd_):
            try:
                import fcntl
                import termios
                cr_ = struct.unpack('hh', fcntl.ioctl(fd_, termios.TIOCGWINSZ, '1234'))
                return cr_
            except:
                pass
        cr = ioctl_gwinsz(0) or ioctl_gwinsz(1) or ioctl_gwinsz(2)
        if not cr:
            try:
                fd = os.open(os.ctermid(), os.O_RDONLY)
                cr = ioctl_gwinsz(fd)
                os.close(fd)
            except:
                pass
        if not cr:
            try:
                cr = (os.environ['LINES'], os.environ['COLUMNS'])
            except:
                return None
        return int(cr[1]), int(cr[0])


def printable(petscii):
    """ convert a petscii string into a printable (unicode or ascii) string """
    global args
    # if in ascii mode (not unicode) we're going to translate non-printable petscii to a question mark
    if 'transascii' not in printable.__dict__:
        printable.transascii = string.maketrans(''.join(chr(x) for x in xrange(0, 0x20)) + ''.join(chr(x) for x in xrange(0x80, 0xa0)), '?'*0x40)
    if args.unicode:
        return cbmdisk.to_unicode(petscii, upper=(args.case == 'upper'), ctrlchrs=True).encode('utf-8')
    else:
        return cbmdisk.to_ascii(petscii.translate(printable.transascii), upper=(args.case == 'upper'))


def lambda_len_bytes(cbm_file):
    try:
        return len(cbm_file.bytes)
    except cbmdisk.error:
        return 0


def human_size(size):
    """ convert a number into a 'human readable' size string """
    magnitude = int(log(size, 2) / 10) if size else 0
    return '{:.1f}{}'.format(float(size) / (1 << (magnitude * 10)), ['', 'K', 'M'][magnitude]).rstrip('0').rstrip('.')


def process_filename(cbm_disk_name):
    """ open a given filename as a cbmdisk.Disk, handle errors """
    try:
        process_disk(cbmdisk.Disk(cbm_disk_name), cbm_disk_name)
    except cbmdisk.error as exception:
        if exception.message['code'] in ('ReadDiskName', 'CreateDisk'):
            # we expect we can open files with supported extensions and report if we can't; for all other files, report it if we aren't suppressing
            if cbm_disk_name.lower().endswith(tuple(cbmdisk.extensions)) or args.quiet < 1:
                print
                print >> sys.stderr, 'cbmdisk.error: \"{}\" [{}]'.format(cbm_disk_name, exception.message['message'])
        else:
            raise


def process_disk(cbm_disk, cbm_disk_name):
    """ list, recursively as needed, the given cbmdisk.Disk or cbmdisk.Files instance """
    global args, colors
    deep_into = []

    disk_name = None
    if isinstance(cbm_disk, cbmdisk.Disk):
        assert isinstance(cbm_disk, cbmdisk.Disk)
        disk_name = cbm_disk.name
        target = cbm_disk
        target.hide_splat_files = not (args.all or args.almost_all)
        files = target.files
    elif isinstance(cbm_disk, cbmdisk.Files):
        assert isinstance(cbm_disk, cbmdisk.Files)
        files = cbm_disk
    else:
        return

    # account for sorting options
    if args.sort_name:
        files = sorted(files, key=attrgetter('name'))
    elif args.sort_size:
        files = sorted(files, key=lambda i: lambda_len_bytes(i))
    elif args.sort_type:
        files = sorted(files, key=attrgetter('type', 'index'))
    else:
        files = list(files)

    # first pass over files to tally up necessary column widths for 'pretty' output
    column_max = {'blocks': 0, 'bytes': 0, 'index': 0, 'track': 0, 'sector': 0}
    column_width = {'blocks': 0, 'bytes': 0, 'splat': 0, 'lock': 0, 'addr': 0, 'type': 0, 'index': 0, 'track': 0, 'sector': 0}
    total_size = 0
    max_name_size = 0
    problem_files = []
    for cbm_file in files:
        assert isinstance(cbm_file, cbmdisk.File)
        if not args.all and (cbm_file.name == chr(0x0)*16 and cbm_file.splat):
            continue
        try:
            column_max['bytes'] = max(len(cbm_file.bytes), column_max['bytes'])
        except:
            problem_files.append(cbm_file.index)
        total_size += cbm_file.size
        column_max['blocks'] = max(cbm_file.size, column_max['blocks'])
        column_max['index'] = max(cbm_file.index, column_max['index'])
        column_max['track'] = max(cbm_file.track, column_max['track'])
        column_max['sector'] = max(cbm_file.sector, column_max['sector'])
        column_width['type'] = 3
        if len(cbm_file.name) > max_name_size:
            max_name_size = len(cbm_file.name)
        if cbm_file.type == 'PRG':
            column_width['addr'] = 5
        if cbm_file.splat:
            column_width['splat'] = 1
        if cbm_file.lock:
            column_width['lock'] = 1
    column_width['blocks'] = len(str(column_max['blocks']))
    column_width['bytes'] = len(str(column_max['bytes']))
    column_width['index'] = len(str(column_max['index']))
    column_width['track'] = len(str(column_max['track']))
    column_width['sector'] = len(str(column_max['sector']))

    # print the full path/file preceding the disk listing
    print cbm_disk_name.decode(sys.getfilesystemencoding()).encode('utf-8') + ':'

    if args.size:
        print 'total {}'.format(total_size)

    # emit the disk name and id if present, lined up above the file list as occurs in a real load"$" listing
    if disk_name:
        printable_disk_name = printable(cbm_disk.name)
        printable_disk_id = printable(cbm_disk.id)
        padding = (column_width['blocks'] + column_width['bytes'] + 2) if args.size else 0
        padding += (column_width['index'] + 1) if args.index else 0
        padding += column_width['track'] + column_width['sector']
        padding += column_width['addr'] + column_width['type'] + column_width['splat'] + column_width['lock']
        print '{0: >{width}}"{1: <16}" {2}'.format(
            '',
            printable_disk_name,
            printable_disk_id,
            width=padding
        )

    cur_column = 0

    # second pass over files to actually output listing
    for cbm_file in files:
        assert isinstance(cbm_file, cbmdisk.File)

        # don't show splats or (likely) unallocated directory entries if not requested
        if not args.all and (cbm_file.name == chr(0x0)*16 and cbm_file.splat):
            continue

        # for a deep list, we'll plunge into subdirs (CBM) and archives
        if args.deep and cbm_file.index not in problem_files and ((cbm_file.type == 'CBM') or cbm_file.is_compressed):
            already_added = False
            # some logic to make sure we deep list a 4zip file only one time
            if cbm_file.name.startswith(('1!', '2!', '3!', '4!')):
                for entry in deep_into:
                    if cbm_file.name[2:] in entry.name:
                        already_added = True
                        break
            if not already_added:
                deep_into.append(cbm_file)

        # prepare the size, if requested
        if args.size:
            size_str = '{color}{0:{width}} '.format(
                cbm_file.size,
                color=Fore.CYAN+Style.BRIGHT if args.color else '',
                width=column_width['blocks']
            )
        else:
            size_str = ''

        # choose an appropriate color, and set aside a load address if applicable
        color = colors[cbm_file.type] if args.color else ''
        addr_str = ''
        if cbm_file.index in problem_files:
            # highlight files that appear corrupt /TODO: deeper analysis: invalid t/s in chain, loop chain, etc
            color = Fore.RED + Back.YELLOW + Style.BRIGHT if args.color else ''
            addr_str = '*****'
        else:
            # highlight prgs w/ BASIC start address
            if (cbm_file.type == "PRG" and len(cbm_file.bytes)) >= 2:
                addr = ord(cbm_file.bytes[0])+ord(cbm_file.bytes[1])*256
                if addr == 0x0801 or addr == 0x1c01:
                    color = Fore.GREEN+Style.BRIGHT if args.color else ''
                addr_str = '${0:04X}'.format(ord(cbm_file.bytes[0])+ord(cbm_file.bytes[1])*256)
            # highlight anything we think is an archive
            if cbm_file.is_compressed:
                color = Fore.RED+Style.BRIGHT if args.color else ''
            # highlight (dim) probably uninteresting del files
            if cbm_file.type == "DEL" and len(cbm_file.bytes) == 0:
                color = Fore.RED+Style.DIM if args.color else ''
            # highlight files we think are GEOSified
            if cbm_file.type == 'USR' and cbm_file.is_geos:
                color = Fore.WHITE+Style.BRIGHT+Back.BLUE if args.color else ''

        # prepare the printable representation of the petscii filename
        printable_file_name = printable(cbm_file.name)

        if args.long:
            # list the file using long form
            print '{blocks}{bytes}{index}{trsr}{addr}{spl}{type}{lck} {0: <16}'.format(
                printable_file_name,
                # printable_file_name.encode('utf-8') if args.unicode else printable_file_name,
                blocks='{color}{: >{width}} '.format(
                    cbm_file.size,
                    color=Fore.CYAN + Style.BRIGHT if args.color else '',
                    width=column_width['blocks']
                ) if args.size else '',
                bytes='{color}{: >{width}} '.format(
                    '' if cbm_file.index in problem_files else human_size(len(cbm_file.bytes)) if args.human_readable else len(cbm_file.bytes),
                    color=Fore.CYAN+Style.DIM if args.color else '',
                    width=column_width['bytes']
                ) if args.size else '',
                index='{color}{: >{width}} '.format(
                    cbm_file.index,
                    color=Fore.WHITE+Style.BRIGHT if args.color else '',
                    width=column_width['index']
                ) if args.index else '',
                trsr='{color}{: >{t_width}},{: <{s_width}} '.format(
                    cbm_file.track,
                    cbm_file.sector,
                    color=Fore.MAGENTA+Style.BRIGHT if args.color else '',
                    t_width=column_width['track'],
                    s_width=column_width['sector']
                ),
                addr='{color}{: >{width}} '.format(
                    addr_str,
                    color=color if args.color else '',
                    width=column_width['addr']
                ),
                spl='*' if cbm_file.splat else ' '*column_width['splat'],
                type=cbm_file.type,
                lck='<' if cbm_file.lock else ' '*column_width['lock']
            )
        else:
            # list the file using short form
            out = size_str + '{type_clr}{0: >{width}} '.format(
                printable_file_name,
                width=max_name_size,
                type_clr=color if args.color else ''
            )
            if not args.one and cur_column + column_width['blocks'] + 1 + max_name_size > term_width:
                print
                cur_column = 0
            print out,
            cur_column += column_width['blocks'] + 1 + max_name_size + 2
            if args.one:
                print

    # terminate the last line of short/column form
    if not args.long and not args.one:
        print
    print

    # deep process all necessary file entries
    for cbm_file in deep_into:
        assert isinstance(cbm_file, cbmdisk.File)
        if cbm_file.is_compressed:
            # deep process an archive file
            try:
                process_disk(cbm_file.uncompress()[0], '{}{}{}'.format(cbm_disk_name, '|', printable(cbm_file.name)))
            except cbmdisk.error as exception:
                # if we can't uncompress a file that we expected to, output an error if allowed
                if exception.message['code'] in 'FileDecompress':
                    if cbm_disk_name.lower().endswith(tuple(cbmdisk.extensions)) and args.quiet < 2:
                        print
                        print >> sys.stderr, '{}{}{}'.format(cbm_disk_name, '|', printable(cbm_file.name)) + ':'
                        print >> sys.stderr, ' {}: [{}]'.format('cbmdisk.error', exception.message['message'])
                else:
                    raise
        elif cbm_file.type == "cbm":
            # deep process a CBM subdirectory
            process_disk(cbm_file.files, '{}{}{}'.format(cbm_disk_name, '|', printable(cbm_file.name)))


if __name__ == '__main__':
    args = cli()

    # if requested, try to find and enable the unicode console module; otherwise ignore unicode argument
    if args.unicode:
        try:
            import win_unicode_console
            win_unicode_console.enable()
        except ImportError:
            args.unicode = False
            sys.exc_clear()

    # we use the terminal width for short form column output
    term_width, term_height = get_terminal_size()

    # if we have colorama module, enable it and setup default colors
    if args.color:
        try:
            import colorama
            from colorama import Back, Fore, Style
            colorama.init(autoreset=True)
            # customize colors here
            colors = {
                'PRG': Style.DIM + Fore.WHITE,
                'SEQ': Style.DIM + Fore.GREEN,
                'REL': Style.BRIGHT + Fore.YELLOW,
                'USR': Style.DIM + Fore.WHITE,
                'DEL': Style.DIM + Fore.YELLOW,
                'CBM': Style.BRIGHT + Fore.BLUE,
                'CMD': Style.BRIGHT + Fore.BLUE,
                '???': Style.DIM + Fore.YELLOW
            }
        except ImportError:
            args.color = False
            sys.exc_clear()

    # glob any path/file arguments as needed
    all_paths = []
    for arg in args.filename:
        if any(i in arg for i in '*?['):
            all_paths += glob.glob(arg)
        else:
            all_paths.append(arg)

    # terminate with error if no path/file matched a glob
    if len(all_paths) == 0:
        print >> sys.stderr, 'error: no matching paths found'
        exit(1)

    # process every globbed path/file argument
    traversed = []
    for path in all_paths:
        # if it's a file, process it directly
        if os.path.isfile(path) and path not in traversed:
            process_filename(path)
            traversed.append(path)
        # if it's a directory and we're recursing, walk it
        elif os.path.isdir(path):
            if args.recursive:
                for walk_root, walk_dirnames, walk_filenames in os.walk(path):
                    if walk_root not in traversed:
                        for walk_filename in walk_filenames:
                            process_filename(os.path.join(walk_root, walk_filename))
                            traversed.append(os.path.join(walk_root, walk_filename))
            else:
                print >> sys.stderr, 'error: {} must be a file, not a directory'.format(path)
                exit(1)
        else:
            print >> sys.stderr, 'error: {} not found'.format(path)
            exit(1)

    exit(0)
