#!/usr/bin/env python
"""
  cbmod targets a selected file on any supported disk, file, or archive
  format and writes its content as a hexidecimal representation to stdout.
"""

# customize your own preferred defaults here, or comment it out
arg_defaults = [('width', 16), ('case', 'upper'), ('unicode', True), ('color', True)]


# standard imports
import argparse
import os
import platform
import string
import sys

# local import for CBMDisk
import cbmdisk


def cli():
    """ handle command line arguments """
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog='using CBMdisk/python v' + cbmdisk.__version__ + ' w/ CBMdisk v' + cbmdisk.library_version
                                     )
    parser.add_argument('path', metavar='FILE',
                        help='a disk/file image or archive filename')
    parser.add_argument('--version', action='version', version='%(prog)s v1.0.0; by Elwix/Style',
                        help='show program version/credits')
    parser.add_argument('-a', '--addr', metavar='ADDR',
                        help='use ADDR as the base address; include first two bytes of FILE in output')
    parser.add_argument('-i', '--index', metavar='INDEX', type=int, required=True,
                        help='dump the file entry at INDEX to stdout')
    parser.add_argument('-w', '--width', metavar='BYTES', type=int, default=16,
                        help='output BYTES bytes per output line')
    parser.add_argument('-C', '--case', choices=['upper', 'lower'], default='lower',
                        help='output as if in upper/graphics or lower/upper mode')
    parser.add_argument('-U', '--unicode', action='store_true', default=False,
                        help='output contents as unicode (from petscii) instead of ascii')
    parser.add_argument('--color', action='store_true', default=True,
                        help='colorize the output')
    try:
        parser.set_defaults(**dict(arg_defaults))
    except NameError:
        pass
    ret_args = parser.parse_args()
    return ret_args


def process_filename(cbm_disk_name):
    """ open a given filename as a cbmdisk.Disk, handle errors """
    try:
        process_disk(cbmdisk.Disk(cbm_disk_name), cbm_disk_name)
    except cbmdisk.error as exception:
        if exception.message['code'] in ('ReadDiskName', 'CreateDisk'):
            # we expect we can open files with supported extensions and report if we can't; for all other files, report it if we aren't suppressing
            if cbm_disk_name.lower().endswith(tuple(cbmdisk.extensions)) or args.quiet < 1:
                print >> sys.stderr, 'cbmdisk.error: \"{}\" [{}]'.format(cbm_disk_name, exception.message['message'])
        else:
            raise


def process_disk(cbm_disk, cbm_disk_name):
    """ find the file on the cbmdisk.Disk at index and write to stdout """
    assert isinstance(cbm_disk, cbmdisk.Disk)
    global args, encoding

    # make sure the indexed file we want actually exists
    if args.index >= len(cbm_disk.files):
        print >> sys.stderr, 'error: \"{}\" has no file at index {} '.format(cbm_disk_name, args.index)
        exit(1)

    # pull the file
    cbm_file = cbm_disk.files[args.index]

    # if no address is given, strip the 2 byte load address and use it
    if not args.addr:
        if len(cbm_file.bytes) > 2:
            base_addr = ord(cbm_file.bytes[0]) + ord(cbm_file.bytes[1]) * 256
            buffer = cbm_file.bytes[2:]
        else:
            print >> sys.stderr, 'error: the file at index {} is too short ({})'.format(args.index, len(cbm_file.bytes))
            exit(1)
    else:
        base_addr = int(args.addr, 0)
        buffer = cbm_file.bytes

    # if in ascii mode (not unicode) we're going to translate non-printable petscii to a question mark
    transtable = string.maketrans(''.join(chr(x) for x in xrange(0, 0x20)) + ''.join(chr(x) for x in xrange(0x80, 0xa0)), '?'*0x40)

    # output the hex and ascii/petscii dump
    line = 0
    cur_column = 0
    printable = ''
    for i in range(0, len(buffer)):
        # output offset if at start of line
        if cur_column == 0:
            print '{}{:0>4X}'.format(
                Fore.GREEN+Style.DIM if args.color else '',
                line * args.width + base_addr
            ),

        # output byte value
        print '{}{:0>2X}'.format(
            Fore.WHITE+Style.BRIGHT if args.color else '',
            ord(buffer[i])
        ),

        # add byte to the printable representation
        if args.unicode:
            printable = printable + cbmdisk.to_unicode(buffer[i], upper=(args.case == 'upper'), ctrlchrs=True)
        else:
            printable = printable + cbmdisk.to_ascii(buffer[i].translate(transtable), upper=(args.case == 'upper'))

        # keep going til we reach the output width, then add the printable representation
        cur_column += 1
        if cur_column == args.width:
            print '{}{}'.format(
                Fore.BLUE+Style.BRIGHT if args.color else '',
                printable.encode('utf-8') if args.unicode else printable
            )
            line += 1
            cur_column = 0
            printable = ''

    # output the final ascii/petscii printable column
    if cur_column > 0:
        print '{}{}'.format(
            Fore.BLUE+Style.BRIGHT if args.color else '',
            ' '*(args.width-cur_column)*3 + (printable.encode('utf-8') if args.unicode else printable)
        )


if __name__ == '__main__':
    args = cli()

    # if requested, try to find and enable the unicode console module; otherwise ignore unicode argument
    if args.unicode:
        try:
            if platform.system() == 'Windows':
                import win_unicode_console
                win_unicode_console.enable()
        except ImportError:
            args.unicode = False
            sys.exc_clear()

    # if we have colorama module, enable it and setup default colors
    if args.color:
        try:
            # imp.find_module('colorama')
            import colorama
            from colorama import Back, Fore, Style
            colorama.init(autoreset=True)
        except ImportError:
            args.color = False
            sys.exc_clear()

    # if the input is a file, process it; reject directories and non-existant input
    if os.path.isfile(args.path):
        process_filename(args.path)
    elif os.path.isdir(args.path):
        print >> sys.stderr, 'error: {} must be a file, not a directory'.format(args.path)
        exit(1)
    else:
        print >> sys.stderr, 'error: {} not found'.format(args.path)
        exit(1)

    exit(0)
